<?php

namespace Drupal\quick_data\Form;

use Drupal\Core\Entity\EntityDisplayRepositoryInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Routing\TrustedRedirectResponse;
use Drupal\Core\Url;
use Drupal\field\FieldConfigInterface;
use Drupal\quick_data\Batch\CsvImportBatch;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\TestTools\Extension\SchemaInspector;

/**
 * Class DefaultForm.
 */
class QuickDataImportForm extends FormBase {

  /**
   * Constructs a \Drupal\demo\Form\Multistep\MultistepFormBase.
   *
   * @var int
   */

  protected $step = 1;

  /**
   * Separator.
   *
   * @var string
   */
  protected $separate = " ; ";

  /**
   * Type Exception.
   *
   * @var string[]
   */
  protected $fieldTypeException = [
    'comment',
    'feeds_item',
    'tablefield',
    'entity_reference_revisions',
  ];

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The entity display repository.
   *
   * @var \Drupal\Core\Entity\EntityDisplayRepositoryInterface
   */
  protected $entityDisplayRepository;

  /**
   * {@inheritdoc}
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, ModuleHandlerInterface $module_handler, EntityFieldManagerInterface $entity_field_manager, EntityDisplayRepositoryInterface $entity_display_repository) {
    $this->entityDisplayRepository = $entity_display_repository;
    $this->entityFieldManager = $entity_field_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->moduleHandler = $module_handler;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('module_handler'),
      $container->get('entity_field.manager'),
      $container->get('entity_display.repository')
    );
  }

  /**
   * {@inheritDoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $type = '', $bundle = '', $field = '', $entity_id = '') {
    $form_state->set('type', $type);
    $form_state->set('bundle', $bundle);
    $form_state->set('field', $field);
    $form_state->set('entity_id', $entity_id);

    if ($this->step != 1) {
      return $this->quickDataImportProcessCsv($form, $form_state);
    }
    return $this->quickDataImportLoadCsv($form, $form_state);
  }

  /**
   * {@inheritDoc}
   */
  public function getFormId() {
    return 'quick_data_import_form';
  }

  /**
   * {@inheritDoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    if (!empty($values["csv_file_upload"])) {
      $form_state->set('csv_file_upload', $form_state->getValue('csv_file_upload'));
      $form_state->set('csv_delimiter', $form_state->getValue('csv_delimiter'));
      $form_state->setRebuild();
      $this->step = 2;
    }
    if (!empty($values["import"])) {
      $store = $form_state->getStorage();
      $file = $this->entityTypeManager->getStorage('file')
        ->load(current($store['csv_file_upload']));
      $filesize = $file->getSize();
      // Big file we will need batch.
      $limitFileSize = 20000;
      if ($filesize > $limitFileSize) {
        $batch = [
          'title' => $this->t('Importing CSV ...'),
          'operations' => $this->operationsBatchLine($file, $store, $form_state->getUserInput()['col']),
          'init_message' => $this->t('Beginning'),
          'progress_message' => $this->t('Processed @current out of @total.'),
          'error_message' => $this->t('An error occurred during processing'),
          'finished' => '\Drupal\quick_data\Batch\CsvImportBatch::csvImportFinished',
        ];
        batch_set($batch);
      }
      else {
        // If not big file import all.
        $data = $this->convertCsvArray($file, $store['csv_delimiter']);
        $num_imported = CsvImportBatch::quickDataImportMapping($store, $form_state->getUserInput()['col'], $data);
        if ($num_imported) {
          $this->messenger()
            ->addStatus($this->t('@count imported items total.', ['@count' => $num_imported]));
        }
      }
      // Redirection! This is because  $form_state['rebuild'] gets set to TRUE.
      if (!empty($dest_url = $this->getRequest()->query->get('destination'))) {
        $parse = parse_url($dest_url);
        if (!empty($parse['host'])) {
          $response = new TrustedRedirectResponse(Url::fromUri($dest_url)
            ->toString());
          $metadata = $response->getCacheableMetadata();
          $metadata->setCacheMaxAge(0);
          $form_state->setResponse($response);
        }
        else {
          $url = Url::fromUri('internal:/' . $dest_url);
        }
      }
      elseif (!empty($store['entity_id'])) {
        $dest_url = implode('/', [$store['bundle'], $store['entity_id']]);
        $url = Url::fromUri('internal:/' . $dest_url);
      }
      if (!empty($url)) {
        $form_state->setRedirectUrl($url);
      }
    }

  }

  /**
   * {@inheritDoc}
   */
  public function quickDataImportLoadCsv($form, &$form_state) {
    $type = $form_state->get('type');
    $bundle = $form_state->get('bundle');
    $field = $form_state->get('field');
    $entity_id = $form_state->get('entity_id');
    // Page one load csv.
    if (!in_array($type, [
      "node",
      "user",
      "taxonomy_term",
      "data",
      "paragraph",
      "field_collection_item",
      "commerce_store",
      "commerce_product",
      "commerce_order",
      "commerce_order_item",
    ])) {
      $this->messenger()->addError($this->t("Your type %type is not available", ['%type' => $bundle]));
      return $form;
    }
    // Check bundle validate.
    $filename = implode('-', [$type, $bundle]) . '.csv';
    if (in_array($type, ['field_collection_item', 'paragraph'])) {
      if (empty($field) || empty($bundle) || empty($entity_id)) {
        $this->messenger()->addError($this->t("Yours variables is not available"));
        return $form;
      }
      $filename = implode('-', [$bundle, $field]) . '.csv';
    }

    $fields_info = $this->fieldsInfo($bundle, $type, $field, $entity_id);
    $options_fields = $this->optionsLoadField($fields_info);
    array_shift($options_fields);
    $form['files']['csv_file_upload'] = [
      '#type' => 'managed_file',
      '#title' => $this->t('Choose a file'),
      '#description' => '<a class="btn btn-success button button--small" id="export" data-filename="' . $filename . '" download="' . $filename . '"> ' . "Download CSV template" . ' <i class="bi bi-filetype-csv"></i> : <br/></a>
<span id="export-header" >' . implode(',', $options_fields) . '</span>
<br/><i>' . $this->t('Use "%separate" for separate multiple values', ['%separate' => $this->separate]) . '</i>
',
      '#required' => TRUE,
      '#upload_validators' => [
        'file_validate_extensions' => ['csv'],
      ],
    ];
    $form['files']['csv_delimiter'] = [
      '#type' => 'select',
      '#title' => $this->t('Character used to separate'),
      '#options' => [
        ',' => ',',
        ';' => ';',
        'tab' => 'Tab',
      ],
    ];

    $form['#attached'] = [
      'library' => ['quick_data/drupal.quick_data'],
      'drupalSettings' => [
        'quick_data' => [
          'entity_type' => $form_state->getValue('type'),
        ],
      ],
    ];
    $form['actions']['#type'] = 'actions';
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Next'),
      '#button_type' => 'primary',
    ];
    return $form;
  }

  /**
   * {@inheritDoc}
   */
  protected function quickDataImportProcessCsv($form, &$form_state) {
    $type = $form_state->get('type');
    $bundle = $form_state->get('bundle');
    $field = $form_state->get('field');
    $entity_id = $form_state->get('entity_id');
    $fields_info = $this->fieldsInfo($bundle, $type, $field, $entity_id);
    $form_state->set('fields_info', $fields_info);
    $options_fields = $this->optionsLoadField($fields_info);
    $field = [
      '#type' => 'select',
      '#options' => $options_fields,
      '#empty_option' => $this->t('- Select -'),
    ];
    $flip_fields = !empty($options_fields) ? array_flip($options_fields) : [];
    $csv_rows = $this->convertCsvArray($form_state->getValue("csv_file_upload"), $form_state->getValue("csv_delimiter"), 10);
    $header = [];
    $header_title = array_shift($csv_rows);
    foreach ($header_title as $col => $title_col) {
      $title_col = trim($title_col);
      $header[$col]["data"] = $field;
      $header[$col]["data"]["#title"] = $title_col;
      $header[$col]["data"]["#name"] = 'col[' . $col . ']';
      if (!empty($flip_fields[$title_col])) {
        $header[$col]["data"]["#value"] = $flip_fields[$title_col];
      }
    }

    $form['table'] = [
      '#theme' => 'table',
      "#title" => $this->t("Mapping field with data"),
      '#header' => $header,
      '#rows' => $csv_rows,
      '#empty' => $this->t('Empty'),
    ];

    $form['actions']['#type'] = 'actions';
    $form['actions']['import'] = [
      '#type' => 'submit',
      '#value' => $this->t('Import'),
      '#button_type' => 'primary',
    ];
    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function convertCsvArray($file, $delimiter = ';', $limit = FALSE) {
    if (is_array($file)) {
      $file = end($file);
    }
    if ($delimiter == 'tab') {
      $delimiter = "\t";
    }
    if (is_numeric($file)) {
      $file = $this->entityTypeManager->getStorage('file')->load($file);
    }
    $filename = $file->get('uri')->value;
    $csv_rows = [];
    if (($f = fopen($filename, "r")) !== FALSE) {
      $i = 0;
      while (($data = fgetcsv($f, 0, $delimiter)) !== FALSE) {
        $csv_rows[] = $data;
        if ($limit && $i++ > $limit) {
          break;
        }
      }
      fclose($f);
    }
    return $csv_rows;
  }

  /**
   * {@inheritDoc}
   */
  private function tableInfo($table) {
    $specification = SchemaInspector::getTablesSpecification($this->moduleHandler, 'quick_data');
    $schema = $specification[$table];
    $infos = [];
    $i = 0;
    foreach ($schema['fields'] as $name => $field) {
      $infos[$name] = [
        'field_name' => $name,
        'label' => !empty($field['description']) ? $field['description'] : $name,
        'type' => $field["type"],
        'columns' => ['value' => ['type' => $field["type"]]],
        'cardinality' => 1,
        'weight' => $i++,
      ];
    }
    return $infos;
  }

  /**
   * {@inheritDoc}
   */
  protected function getTarget($field_definition) {
    $propertyDefinitions = $field_definition->getPropertyDefinitions();
    $type = $field_definition->getType();
    $targetDefinition = [
      'text_with_summary' => ['value', 'summary'],
      'link' => ['uri', 'title'],
      'address' => [
        'organization',
        'given_name',
        'family_name',
        'country_code',
        'address_line1',
        'address_line2',
        'postal_code',
        'locality',
        // 'langcode',
        // 'administrative_area',
        // 'dependent_locality',
        // 'additional_name',
        // 'sorting_code',
      ],
      'name' => [
        'title',
        'given',
        'middle',
        'family',
        // 'generational',
        // 'credentials',
      ],
      'daterange' => ['value', 'end_value'],
      'entity_reference' => ['target_id'],
      'image' => ['target_id', 'alt', 'title'],
      'double_field' => ['first', 'second'],
      'triples_field' => ['first', 'second', 'third'],
    ];
    $target = [];
    if ($type == 'data_field') {
      foreach ($propertyDefinitions as $column => $proDef) {
        $target[$column] = (string) $proDef->getLabel();
      }
    }
    if (!empty($targetDefinition[$type])) {
      foreach ($targetDefinition[$type] as $column) {
        if (!empty($propertyDefinitions[$column])) {
          $target[$column] = (string) $propertyDefinitions[$column]->getLabel();
        }
      }
    }
    return $target;
  }

  /**
   * {@inheritDoc}
   */
  private function fieldsInfo($bundle, $type, $field = '', $entity_id = FALSE) {
    if ($bundle == 'data') {
      return $this->tableInfo($type);
    }
    if ($type == 'paragraph') {
      $content_type = $this->entityTypeManager->getStorage($bundle)->load($entity_id)->getType();
      $field_instance = $this->entityTypeManager->getStorage('field_config')
        ->load(implode('.', [$bundle, $content_type, $field]));
      $targetBundle = array_key_first($field_instance->getSetting("handler_settings")["target_bundles"]);
      $field_definitions = $this->entityTypeManager->getStorage($type)->create(['type' => $targetBundle])->getFieldDefinitions();
    }
    else {
      $field_definitions = $this->entityFieldManager->getFieldDefinitions($type, $bundle);
    }

    foreach ($field_definitions as $name => $definition) {
      if ($definition instanceof FieldConfigInterface) {
        if (empty($field_type) || $definition->getType() == $field_type) {
          $fields[$name] = $definition;
        }
      }
    }
    $infos = [];
    $formDisplay = $this->entityDisplayRepository->getFormDisplay($type, $bundle, 'default');
    if (!empty($fields)) {
      foreach ($fields as $field) {
        $field_type = $field->getType();
        $field_info = $field->getFieldStorageDefinition();
        $module = $field_info->get('module');
        if (in_array($field_type, $this->fieldTypeException)) {
          continue;
        }
        $name = $field->getName();
        $fieldComponent = $formDisplay->getComponent($name);
        $infos[$name] = [
          'field_name' => $name,
          'label' => $field->getLabel(),
          'type' => $field->getType(),
          'widget' => $field_info->getType(),
          'columns' => $this->getTarget($field_info),
          'cardinality' => $field_info->getCardinality(),
          'allowed_values' => options_allowed_values($field_info),
          'weight' => $fieldComponent["weight"] ?? 0,
          'module' => $module,
          'schema' => $field_info->getSchema()['columns'],
          'settings' => $field->getSettings(),
        ];
      }
    }
    switch ($type) {
      case 'node':
        $infos['status'] = [
          'field_name' => 'status',
          'label' => $this->t('Status'),
          'type' => 'status',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        $infos['title'] = [
          'field_name' => 'title',
          'label' => $this->t('Title'),
          'type' => 'title',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        break;

      case 'user':
        $infos['status'] = [
          'field_name' => 'status',
          'label' => $this->t('Status'),
          'type' => 'status',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        $infos['mail'] = [
          'field_name' => 'mail',
          'label' => $this->t('Email'),
          'type' => 'mail',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        $infos['name'] = [
          'field_name' => 'name',
          'label' => $this->t('Name'),
          'type' => 'name',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        break;

      case 'taxonomy_term':
        $infos['description'] = [
          'field_name' => 'description',
          'label' => $this->t('description'),
          'type' => 'description',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        $infos['name'] = [
          'field_name' => 'name',
          'label' => $this->t('Name'),
          'type' => 'name',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        break;

      case 'commerce_store':
        $field_address = $field_definitions["address"];
        $field_info_address = $field_address->getFieldStorageDefinition();
        $infos['address'] = [
          'field_name' => 'address',
          'label' => $field_address->getLabel(),
          'type' => $field_address->getType(),
          'widget' => $field_info_address->getType(),
          'cardinality' => 1,
          'allowed_values' => options_allowed_values($field_info_address),
          'weight' => $fieldComponent["weight"] ?? 0,
          'module' => $field_info_address->get('module'),
          'schema' => $field_info_address->getSchema()['columns'],
          'settings' => $field_address->getSettings(),
          'columns' => $this->getTarget($field_info_address),
        ];
        $infos['mail'] = [
          'field_name' => 'mail',
          'label' => $this->t('Email'),
          'type' => 'mail',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        $infos['name'] = [
          'field_name' => 'name',
          'label' => $this->t('Name'),
          'type' => 'name',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        break;

      case 'commerce_product':
        $infos['stores'] = [
          'field_name' => 'stores',
          'label' => $this->t('Store'),
          'type' => 'store',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        $infos['status'] = [
          'field_name' => 'status',
          'label' => $this->t('Status'),
          'type' => 'status',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        $infos['title'] = [
          'field_name' => 'title',
          'label' => $this->t('Title'),
          'type' => 'title',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        $field_variations = $field_definitions["variations"];
        $field_info_variations = $field_variations->getFieldStorageDefinition();
        $infos['variations'] = [
          'field_name' => 'variations',
          'label' => $field_variations->getLabel(),
          'type' => $field_variations->getType(),
          'widget' => $field_info_variations->getType(),
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        break;

      case 'commerce_order':
        $infos['mail'] = [
          'field_name' => 'mail',
          'label' => $this->t('Email'),
          'type' => 'mail',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        $infos['status'] = [
          'field_name' => 'status',
          'label' => $this->t('Status'),
          'type' => 'status',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        $infos['order_number'] = [
          'field_name' => 'order_number',
          'label' => $this->t('Order number'),
          'type' => 'string',
          'widget' => 'text_textfield',
          'module' => $bundle,
          'cardinality' => 1,
          'weight' => 1,
        ];
        break;

      case 'commerce_order_item':
        $order_item_fields = [
          'order_id',
          'purchased_entity',
          'title',
          'quantity',
          'unit_price',
          'overridden_unit_price',
          'total_price',
          'adjustments',
        ];
        foreach ($order_item_fields as $item_field) {
          $field_import = $field_definitions[$item_field];
          $field_info_import = $field_import->getFieldStorageDefinition();
          $infos[$item_field] = [
            'field_name' => $item_field,
            'label' => $field_import->getLabel(),
            'type' => $field_import->getType(),
            'widget' => $field_info_import->getType(),
            'module' => $bundle,
            'cardinality' => 1,
            'weight' => 1,
          ];
        }
        break;
    }

    return $infos;
  }

  /**
   * {@inheritDoc}
   */
  private function optionsLoadField($infos) {
    uasort($infos, 'Drupal\Component\Utility\SortArray::sortByWeightElement');
    $options = ['_none' => (string) $this->t('- None -')];
    foreach ($infos as $field_name => $info) {
      if (in_array($info['type'], $this->fieldTypeException)) {
        continue;
      }
      if (!empty($info['columns'])) {
        foreach ($info['columns'] as $name => $label) {
          $options["$field_name:$name"] = trim((string) $info['label'] . ': ' . strip_tags($label));
        }
        continue;
      }
      $options[$field_name] = (string) $info['label'];
    }
    return $options;
  }

  /**
   * Operation Batch import.
   *
   * {@inheritdoc}
   */
  private function operationsBatchLine($file, $store, $mapping) {
    $filename = $file->get('uri')->value;
    $batch = [];
    if ($f = fopen($filename, "r")) {
      $row_index = 0;
      if ($store['csv_delimiter'] == 'tab') {
        $store['csv_delimiter'] = "\t";
      }
      while (($line = fgetcsv($f, 0, $store['csv_delimiter'])) !== FALSE) {
        // Remove first line header csv.
        if (!$row_index++) {
          continue;
        }
        $batch[] = [
          '\Drupal\quick_data\Batch\CsvImportBatch::csvImportLine',
          [[$line], $store, $mapping],
        ];
      }
      fclose($f);
    }
    return $batch;
  }

}
